@@ -1,4 +1,4 @@
-# Copyright (C) 2017-2022 MariaDB
+# Copyright (C) 2017-2024 MariaDB
 # Copyright (C) 2012-2015 Codership Oy
 #
 # This program is free software; you can redistribute it and/or modify
@@ -15,13 +15,69 @@
 # Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston
 # MA  02110-1335  USA.
 
-# This is a common command line parser to be sourced by other SST scripts
+# This is a common command line parser and common functions to
+# be sourced by other SST scripts.
 
 trap 'exit 32' HUP PIPE
 trap 'exit 3'  INT QUIT TERM
 
-# Setting the path for some utilities on CentOS
-export PATH="$PATH:/usr/sbin:/usr/bin:/sbin:/bin"
+OS="$(uname)"
+
+# Setting the paths for some utilities on CentOS
+export PATH="${PATH:+$PATH:}/usr/local/bin:/usr/local/sbin:/usr/bin:/bin:/usr/sbin:/sbin"
+if [ "$OS" != 'Darwin' ]; then
+    export LD_LIBRARY_PATH="${LD_LIBRARY_PATH:+$LD_LIBRARY_PATH:}/usr/local/lib:/usr/lib:/lib:/opt/lib"
+fi
+
+commandex()
+{
+    if [ -n "$BASH_VERSION" ]; then
+        command -v "$1" || :
+    elif [ -x "$1" ]; then
+        echo "$1"
+    else
+        which "$1" || :
+    fi
+}
+
+with_bash_42=0
+with_printf=1
+if [ -z "$BASH_VERSION" ]; then
+    [ -z "$(commandex printf)" ] && with_printf=0
+else
+    [ "${BASH_VERSINFO[0]}" -eq 4 -a "${BASH_VERSINFO[1]}" -ge 2 -o \
+      "${BASH_VERSINFO[0]}" -gt 4 ] && with_bash_42=1
+fi
+
+wsrep_log()
+{
+    local t
+    # echo everything to stderr so that it gets into common error log
+    # deliberately made to look different from the rest of the log
+    if [ "$OS" = 'Linux' ]; then
+        t=$(date '+%Y%m%d %H:%M:%S.%3N')
+    elif [ $with_bash_42 -ne 0 ]; then
+        printf -v t '%(%Y%m%d %H:%M:%S)T.000'
+    else
+        t=$(date '+%Y%m%d %H:%M:%S.000')
+    fi
+    echo "WSREP_SST: $* ($t)" >&2
+}
+
+wsrep_log_error()
+{
+    wsrep_log "[ERROR] $*"
+}
+
+wsrep_log_warning()
+{
+    wsrep_log "[WARNING] $*"
+}
+
+wsrep_log_info()
+{
+    wsrep_log "[INFO] $*"
+}
 
 trim_string()
 {
@@ -35,9 +91,9 @@ trim_string()
             y=${#y}
             x=$(( z-x-1 ))
             y=$(( y-x+1 ))
-            printf '%s' "${1:$x:$y}"
+            echo "${1:$x:$y}"
         else
-            printf ''
+            echo ''
         fi
     else
         local pattern="[[:space:]${2:-}]"
@@ -47,18 +103,70 @@ trim_string()
 
 trim_dir()
 {
-    local t=$(trim_string "$1")
-    if [ "$t" != '/' ]; then
-        if [ "${t%/}" != "$t" ]; then
-            t=$(trim_string "${t%/}")
+    if [ -n "$BASH_VERSION" ]; then
+        local pattern="![:space:]${2:-}"
+        local x="${1#*[$pattern]}"
+        local z=${#1}
+        x=${#x}
+        if [ $x -ne $z ]; then
+            local y="${1%[$pattern/]*}"
+            y=${#y}
+            x=$(( z-x-1 ))
+            y=$(( y-x+1 ))
+            x="${1:$x:$y}"
+            [ -z "$x" ] && x='.'
+            echo "$x"
+        else
+            echo ''
         fi
     else
-        t='.'
+        local pattern="[:space:]${2:-}"
+        local x=$(echo "$1" | sed -E "s/^[$pattern]+|[$pattern/]+\$//g")
+        if [ -n "$x" ]; then
+            echo "$x"
+        elif "${1#*/}" != "$1"; then
+            echo '.'
+        else
+            echo ''
+        fi
     fi
+}
+
+trim_right()
+{
     if [ -n "$BASH_VERSION" ]; then
-        printf '%s' "$t"
+        local pattern="[![:space:]${2:-}]"
+        local z=${#1}
+        local y="${1%$pattern*}"
+        y=${#y}
+        if [ $y -ne $z ]; then
+            y=$(( y+1 ))
+            echo "${1:0:$y}"
+        else
+            echo ''
+        fi
     else
-        echo "$t"
+        local pattern="[[:space:]${2:-}]"
+        echo "$1" | sed -E "s/$pattern+\$//g"
+    fi
+}
+
+trim_left()
+{
+    if [ -n "$BASH_VERSION" ]; then
+        local pattern="[![:space:]${2:-}]"
+        local x="${1#*$pattern}"
+        local z=${#1}
+        x=${#x}
+        if [ $x -ne $z ]; then
+            x=$(( z-x-1 ))
+            echo "${1:$x:$z}"
+        else
+            echo ''
+        fi
+    else
+        local pattern="[[:space:]${2:-}]"
+        echo "$1" | sed -E "s/^$pattern+//g"
     fi
 }
 
@@ -72,11 +180,7 @@ to_minuses()
        x="$t"
        t="${t#*_}"
     done
-    if [ -n "$BASH_VERSION" ]; then
-        printf '%s' "$r$x"
-    else
-        echo "$r$x"
-    fi
+    echo "$r$x"
 }
 
 WSREP_SST_OPT_BYPASS=0
@@ -88,7 +192,6 @@ WSREP_SST_OPT_DATA=""
 WSREP_SST_OPT_AUTH="${WSREP_SST_OPT_AUTH:-}"
 WSREP_SST_OPT_USER="${WSREP_SST_OPT_USER:-}"
 WSREP_SST_OPT_PSWD="${WSREP_SST_OPT_PSWD:-}"
-WSREP_SST_OPT_REMOTE_AUTH="${WSREP_SST_OPT_REMOTE_AUTH:-}"
 WSREP_SST_OPT_DEFAULT=""
 WSREP_SST_OPT_DEFAULTS=""
 WSREP_SST_OPT_EXTRA_DEFAULT=""
@@ -101,17 +204,19 @@ WSREP_SST_OPT_ADDR=""
 WSREP_SST_OPT_ADDR_PORT=""
 WSREP_SST_OPT_HOST=""
 WSREP_SST_OPT_HOST_UNESCAPED=""
+ARIA_LOG_DIR=""
 INNODB_DATA_HOME_DIR=$(trim_dir "${INNODB_DATA_HOME_DIR:-}")
 INNODB_LOG_GROUP_HOME=$(trim_dir "${INNODB_LOG_GROUP_HOME:-}")
 INNODB_UNDO_DIR=$(trim_dir "${INNODB_UNDO_DIR:-}")
 INNODB_BUFFER_POOL=""
+INNODB_BUFFER_POOL_SIZE=""
 INNODB_FORCE_RECOVERY=""
 INNOEXTRA=""
 
 while [ $# -gt 0 ]; do
 case "$1" in
     '--address')
-        WSREP_SST_OPT_ADDR="$2"
+        WSREP_SST_OPT_ADDR=$(trim_string "$2")
         #
         # Break address string into host:port/path parts
         #
@@ -119,20 +224,22 @@ case "$1" in
         \[*)
             # IPv6
             # Remove the starting and ending square brackets, if present:
-            addr_no_bracket="${WSREP_SST_OPT_ADDR#\[}"
+            addr="${WSREP_SST_OPT_ADDR#\[}"
+            addr=$(trim_right "${addr%%\]*}")
             # Some utilities and subsequent code require an address
             # without square brackets:
-            readonly WSREP_SST_OPT_HOST_UNESCAPED="${addr_no_bracket%%\]*}"
+            readonly WSREP_SST_OPT_HOST_UNESCAPED="$addr"
             # Square brackets are needed in most cases:
-            readonly WSREP_SST_OPT_HOST="[$WSREP_SST_OPT_HOST_UNESCAPED]"
+            readonly WSREP_SST_OPT_HOST="[$addr]"
             # Mark this address as IPv6:
             readonly WSREP_SST_OPT_HOST_IPv6=1
             # Let's remove the leading part that contains the host address:
             remain="${WSREP_SST_OPT_ADDR#*\]}"
             ;;
         *)
-            readonly WSREP_SST_OPT_HOST="${WSREP_SST_OPT_ADDR%%[:/]*}"
-            readonly WSREP_SST_OPT_HOST_UNESCAPED="$WSREP_SST_OPT_HOST"
+            addr=$(trim_right "${WSREP_SST_OPT_ADDR%%[:/]*}")
+            readonly WSREP_SST_OPT_HOST="$addr"
+            readonly WSREP_SST_OPT_HOST_UNESCAPED="$addr"
             readonly WSREP_SST_OPT_HOST_IPv6=0
             # Let's remove the leading part that contains the host address:
             remain="${WSREP_SST_OPT_ADDR#*[:/]}"
@@ -154,17 +261,18 @@ case "$1" in
         else
             readonly WSREP_SST_OPT_PATH=""
         fi
+        WSREP_SST_OPT_ADDR_PORT=$(trim_right "$WSREP_SST_OPT_ADDR_PORT")
         # Remove the module name part from the string, which ends with "/":
         remain="${WSREP_SST_OPT_PATH#*/}"
         # This operation removes the tail after the very first occurrence
         # of the "/" character, inclusively:
-        readonly WSREP_SST_OPT_MODULE="${WSREP_SST_OPT_PATH%%/*}"
+        readonly WSREP_SST_OPT_MODULE=$(trim_right "${WSREP_SST_OPT_PATH%%/*}")
         # If there is one more "/" in the string, then everything before
         # it will be the LSN, otherwise the LSN is empty:
         if [ "$remain" != "$WSREP_SST_OPT_PATH" ]; then
             # Extract the part that matches the LSN by removing all
             # characters starting from the very first "/":
-            readonly WSREP_SST_OPT_LSN="${remain%%/*}"
+            readonly WSREP_SST_OPT_LSN=$(trim_right "${remain%%/*}")
             # Exctract everything after the first occurrence of
             # the "/" character in the string:
             source="$remain"
@@ -176,7 +284,7 @@ case "$1" in
                 # Let's extract the version number by removing the tail
                 # after the very first occurence of the "/" character
                 # (inclusively):
-                readonly WSREP_SST_OPT_SST_VER="${remain%%/*}"
+                readonly WSREP_SST_OPT_SST_VER=$(trim_right "${remain%%/*}")
             else
                 readonly WSREP_SST_OPT_SST_VER=""
             fi
@@ -198,6 +306,11 @@ case "$1" in
         readonly WSREP_SST_OPT_DATA=$(trim_dir "$2")
         shift
         ;;
+    '--aria-log-dir-path')
+        # Let's remove the trailing slash:
+        readonly ARIA_LOG_DIR=$(trim_dir "$2")
+        shift
+        ;;
     '--innodb-data-home-dir')
         # Let's remove the trailing slash:
         readonly INNODB_DATA_HOME_DIR=$(trim_dir "$2")
@@ -217,42 +330,51 @@ case "$1" in
         readonly INNODB_BUFFER_POOL=$(trim_string "$2")
         shift
         ;;
+    '--innodb-buffer-pool-size')
+        readonly INNODB_BUFFER_POOL_SIZE=$(trim_string "$2")
+        shift
+        ;;
     '--defaults-file')
-        readonly WSREP_SST_OPT_DEFAULT="$1=$2"
-        readonly WSREP_SST_OPT_DEFAULTS="$1='$2'"
+        file=$(trim_string "$2")
+        readonly WSREP_SST_OPT_DEFAULT="$1=$file"
+        readonly WSREP_SST_OPT_DEFAULTS="$1='$file'"
         shift
         ;;
     '--defaults-extra-file')
-        readonly WSREP_SST_OPT_EXTRA_DEFAULT="$1=$2"
-        readonly WSREP_SST_OPT_EXTRA_DEFAULTS="$1='$2'"
+        file=$(trim_string "$2")
+        readonly WSREP_SST_OPT_EXTRA_DEFAULT="$1=$file"
+        readonly WSREP_SST_OPT_EXTRA_DEFAULTS="$1='$file'"
         shift
         ;;
     '--defaults-group-suffix')
-        readonly WSREP_SST_OPT_SUFFIX_DEFAULT="$1=$2"
-        readonly WSREP_SST_OPT_SUFFIX_VALUE="$2"
+        suffix=$(trim_string "$2")
+        readonly WSREP_SST_OPT_SUFFIX_DEFAULT="$1=$suffix"
+        readonly WSREP_SST_OPT_SUFFIX_VALUE="$suffix"
         shift
         ;;
     '--host')
-        case "$2" in
+        addr=$(trim_string "$2")
+        case "$addr" in
         \[*)
             # IPv6
             # Remove the starting and ending square brackets, if present:
-            addr_no_bracket="${2#\[}"
+            addr="${addr#\[}"
+            addr=$(trim_right "${addr%%\]*}")
             # Some utilities and subsequent code require an address
             # without square brackets:
-            readonly WSREP_SST_OPT_HOST_UNESCAPED="${addr_no_bracket%%\]*}"
+            readonly WSREP_SST_OPT_HOST_UNESCAPED="$addr"
             # Square brackets are needed in most cases:
-            readonly WSREP_SST_OPT_HOST="[${WSREP_SST_OPT_HOST_UNESCAPED}]"
+            readonly WSREP_SST_OPT_HOST="[$addr]"
             # Mark this address as IPv6:
             readonly WSREP_SST_OPT_HOST_IPv6=1
             ;;
         *)
-            readonly WSREP_SST_OPT_HOST="$2"
-            readonly WSREP_SST_OPT_HOST_UNESCAPED="$2"
+            readonly WSREP_SST_OPT_HOST="$addr"
+            readonly WSREP_SST_OPT_HOST_UNESCAPED="$addr"
             readonly WSREP_SST_OPT_HOST_IPv6=0
             ;;
         esac
-        WSREP_SST_OPT_ADDR="$WSREP_SST_OPT_HOST"
+        WSREP_SST_OPT_ADDR="$addr"
         shift
         ;;
     '--local-port')
@@ -272,11 +394,11 @@ case "$1" in
         shift
         ;;
     '--role')
-        readonly WSREP_SST_OPT_ROLE="$2"
+        readonly WSREP_SST_OPT_ROLE=$(trim_string "$2")
         shift
         ;;
     '--socket')
-        readonly WSREP_SST_OPT_SOCKET="$2"
+        readonly WSREP_SST_OPT_SOCKET=$(trim_string "$2")
         shift
         ;;
     '--user')
@@ -284,23 +406,23 @@ case "$1" in
         shift
         ;;
     '--gtid')
-        readonly WSREP_SST_OPT_GTID="$2"
+        readonly WSREP_SST_OPT_GTID=$(trim_string "$2")
         shift
         ;;
     '--binlog'|'--log-bin')
-        readonly WSREP_SST_OPT_BINLOG="$2"
+        readonly WSREP_SST_OPT_BINLOG=$(trim_string "$2")
         shift
         ;;
     '--binlog-index'|'--log-bin-index')
-        WSREP_SST_OPT_BINLOG_INDEX="$2"
+        WSREP_SST_OPT_BINLOG_INDEX=$(trim_string "$2")
         shift
         ;;
     '--log-basename')
-        readonly WSREP_SST_OPT_LOG_BASENAME="$2"
+        readonly WSREP_SST_OPT_LOG_BASENAME=$(trim_string "$2")
         shift
         ;;
     '--gtid-domain-id')
-        readonly WSREP_SST_OPT_GTID_DOMAIN_ID="$2"
+        readonly WSREP_SST_OPT_GTID_DOMAIN_ID=$(trim_string "$2")
         shift
         ;;
     '--mysqld-args')
@@ -336,9 +458,11 @@ case "$1" in
                        # option name:
                        if [ -n "$BASH_VERSION" ]; then
                            option="${options:0:1}"
+                       elif [ $with_printf -ne 0 ]; then
+                           option=$(printf '%.1s' "$options")
                        else
-                           # If it's not bash, then we need to use slow
-                           # external utilities:
+                           # If it's not bash and without printf,
+                           # then we need to use slow external utilities:
                            option=$(echo "$options" | cut -c1)
                        fi
                        # And the subsequent characters consider option value:
@@ -458,6 +582,12 @@ case "$1" in
                # from mysqld's argument list:
                skip_mysqld_arg=0
                case "$option" in
+                   '--aria-log-dir-path')
+                       if [ -z "$ARIA_LOG_DIR" ]; then
+                           MYSQLD_OPT_ARIA_LOG_DIR=$(trim_dir "$value")
+                       fi
+                       skip_mysqld_arg=1
+                       ;;
                    '--innodb-data-home-dir')
                        if [ -z "$INNODB_DATA_HOME_DIR" ]; then
                            MYSQLD_OPT_INNODB_DATA_HOME_DIR=$(trim_dir "$value")
@@ -482,6 +612,12 @@ case "$1" in
                        fi
                        skip_mysqld_arg=1
                        ;;
+                   '--innodb-buffer-pool-size')
+                       if [ -z "$INNODB_BUFFER_POOL_SIZE" ]; then
+                           MYSQLD_OPT_INNODB_BUFFER_POOL_SIZE=$(trim_string "$value")
+                       fi
+                       skip_mysqld_arg=1
+                       ;;
                    '--innodb-force-recovery')
                        if [ -n "$value" -a "$value" != "0" ]; then
                            INNODB_FORCE_RECOVERY=$(trim_string "$value")
@@ -551,6 +687,10 @@ readonly WSREP_SST_OPT_PROGRESS
 
 # The same argument can be present on the command line several
 # times, in this case we must take its last value:
+if [ -n "${MYSQLD_OPT_ARIA_LOG_DIR:-}" -a \
+     -z "$ARIA_LOG_DIR" ]; then
+    readonly ARIA_LOG_DIR="$MYSQLD_OPT_ARIA_LOG_DIR"
+fi
 if [ -n "${MYSQLD_OPT_INNODB_DATA_HOME_DIR:-}" -a \
      -z "$INNODB_DATA_HOME_DIR" ]; then
     readonly INNODB_DATA_HOME_DIR="$MYSQLD_OPT_INNODB_DATA_HOME_DIR"
@@ -567,6 +707,10 @@ if [ -n "${MYSQLD_OPT_INNODB_BUFFER_POOL:-}" -a \
      -z "$INNODB_BUFFER_POOL" ]; then
     readonly INNODB_BUFFER_POOL="$MYSQLD_OPT_INNODB_BUFFER_POOL"
 fi
+if [ -n "${MYSQLD_OPT_INNODB_BUFFER_POOL_SIZE:-}" -a \
+     -z "$INNODB_BUFFER_POOL_SIZE" ]; then
+    readonly INNODB_BUFFER_POOL_SIZE="$MYSQLD_OPT_INNODB_BUFFER_POOL_SIZE"
+fi
 if [ -n "${MYSQLD_OPT_LOG_BIN:-}" -a \
      -z "$WSREP_SST_OPT_BINLOG" ]; then
     readonly WSREP_SST_OPT_BINLOG="$MYSQLD_OPT_LOG_BIN"
@@ -608,6 +752,9 @@ if [ -n "$WSREP_SST_OPT_LOG_BASENAME" ]; then
         WSREP_SST_OPT_MYSQLD="--log-basename='$WSREP_SST_OPT_LOG_BASENAME'"
     fi
 fi
+if [ -n "$ARIA_LOG_DIR" ]; then
+    INNOEXTRA="$INNOEXTRA --aria-log-dir-path='$ARIA_LOG_DIR'"
+fi
 if [ -n "$INNODB_DATA_HOME_DIR" ]; then
     INNOEXTRA="$INNOEXTRA --innodb-data-home-dir='$INNODB_DATA_HOME_DIR'"
 fi
@@ -620,6 +767,9 @@ fi
 if [ -n "$INNODB_BUFFER_POOL" ]; then
     INNOEXTRA="$INNOEXTRA --innodb-buffer-pool-filename='$INNODB_BUFFER_POOL'"
 fi
+if [ -n "$INNODB_BUFFER_POOL_SIZE" ]; then
+    INNOEXTRA="$INNOEXTRA --innodb-buffer-pool-size='$INNODB_BUFFER_POOL_SIZE'"
+fi
 if [ -n "$WSREP_SST_OPT_BINLOG" ]; then
     INNOEXTRA="$INNOEXTRA --log-bin='$WSREP_SST_OPT_BINLOG'"
     if [ -n "$WSREP_SST_OPT_BINLOG_INDEX" ]; then
@@ -728,58 +878,11 @@ WSREP_SST_OPT_ADDR="$WSREP_SST_OPT_HOST:$WSREP_SST_OPT_PORT$sst_path"
 readonly WSREP_SST_OPT_ADDR
 readonly WSREP_SST_OPT_ADDR_PORT
 
-commandex()
-{
-    if [ -n "$BASH_VERSION" ]; then
-        command -v "$1" || :
-    elif [ -x "$1" ]; then
-        echo "$1"
-    else
-        which "$1" || :
-    fi
-}
-
 # try to use my_print_defaults, mysql and mysqldump that come
 # with the sources (for MTR suite):
 script_binary=$(dirname "$0")
 SCRIPTS_DIR=$(cd "$script_binary"; pwd)
 EXTRA_DIR="$SCRIPTS_DIR/../extra"
-CLIENT_DIR="$SCRIPTS_DIR/../client"
-
-if [ -x "$CLIENT_DIR/mariadb" ]; then
-    MYSQL_CLIENT="$CLIENT_DIR/mariadb"
-else
-    MYSQL_CLIENT=$(commandex 'mariadb')
-fi
-
-if [ -x "$CLIENT_DIR/mariadb-dump" ]; then
-    MYSQLDUMP="$CLIENT_DIR/mariadb-dump"
-else
-    MYSQLDUMP=$(commandex 'mariadb-dump')
-fi
-
-wsrep_log()
-{
-    # echo everything to stderr so that it gets into common error log
-    # deliberately made to look different from the rest of the log
-    local readonly tst=$(date "+%Y%m%d %H:%M:%S.%N" | cut -b -21)
-    echo "WSREP_SST: $* ($tst)" >&2
-}
-
-wsrep_log_error()
-{
-    wsrep_log "[ERROR] $*"
-}
-
-wsrep_log_warning()
-{
-    wsrep_log "[WARNING] $*"
-}
-
-wsrep_log_info()
-{
-    wsrep_log "[INFO] $*"
-}
 
 if [ -x "$SCRIPTS_DIR/my_print_defaults" ]; then
     MY_PRINT_DEFAULTS="$SCRIPTS_DIR/my_print_defaults"
@@ -869,11 +972,7 @@ parse_cnf()
     # Truncate spaces:
     [ -n "$reval" ] && reval=$(trim_string "$reval")
 
-    if [ -n "$BASH_VERSION" ]; then
-        printf '%s' "$reval"
-    else
-        echo "$reval"
-    fi
+    echo "$reval"
 }
 
 #
@@ -926,16 +1025,8 @@ in_config()
             break
         fi
     done
-    if [ -n "$BASH_VERSION" ]; then
-        printf '%s' $found
-    else
-        echo $found
-    fi
-}
 
-wsrep_auth_not_set()
-{
-    [ -z "$WSREP_SST_OPT_AUTH" ]
+    echo $found
 }
 
 # Get rid of incorrect values resulting from substitution
@@ -953,12 +1044,12 @@ fi
 # Let's read the value of the authentication string from the
 # configuration file so that it does not go to the command line
 # and does not appear in the ps output:
-if wsrep_auth_not_set; then
+if [ -z "$WSREP_SST_OPT_AUTH" ]; then
     WSREP_SST_OPT_AUTH=$(parse_cnf 'sst' 'wsrep-sst-auth')
 fi
 
 # Splitting WSREP_SST_OPT_AUTH as "user:password" pair:
-if ! wsrep_auth_not_set; then
+if [ -n "$WSREP_SST_OPT_AUTH" ]; then
     # Extract username as shortest prefix up to first ':' character:
     WSREP_SST_OPT_AUTH_USER="${WSREP_SST_OPT_AUTH%%:*}"
     if [ -z "$WSREP_SST_OPT_USER" ]; then
@@ -982,19 +1073,57 @@ if ! wsrep_auth_not_set; then
     fi
 fi
 
+WSREP_SST_OPT_REMOTE_AUTH="${WSREP_SST_OPT_REMOTE_AUTH:-}"
+WSREP_SST_OPT_REMOTE_USER=
+WSREP_SST_OPT_REMOTE_PSWD=
+if [ -n "$WSREP_SST_OPT_REMOTE_AUTH" ]; then
+    # Split auth string at the last ':'
+    WSREP_SST_OPT_REMOTE_USER="${WSREP_SST_OPT_REMOTE_AUTH%%:*}"
+    WSREP_SST_OPT_REMOTE_PSWD="${WSREP_SST_OPT_REMOTE_AUTH#*:}"
+fi
+
+# Reads incoming data from STDIN and sets the variables
+#
+# Globals:
+#   WSREP_SST_OPT_USER (sets this variable)
+#   WSREP_SST_OPT_PSWD (sets this variable)
+#
+# Parameters:
+#   None
+#
+read_variables_from_stdin()
+{
+    while read line; do
+        local key="${line%%=*}"
+        local value=""
+        [ "$key" != "$line" ] && value="${line#*=}"
+        case "$key" in
+            'sst_user')
+                WSREP_SST_OPT_USER="$value"
+                ;;
+            'sst_password')
+                WSREP_SST_OPT_PSWD="$value"
+                ;;
+            'sst_remote_user')
+                WSREP_SST_OPT_REMOTE_USER="$value"
+                ;;
+            'sst_remote_password')
+                WSREP_SST_OPT_REMOTE_PSWD="$value"
+                ;;
+            *)
+                wsrep_log_warning "Unrecognized input: $line"
+        esac
+    done
+    return 0
+}
+
+[ "$WSREP_SST_OPT_ROLE" = "donor" ] && read_variables_from_stdin || :
+
 readonly WSREP_SST_OPT_USER
 readonly WSREP_SST_OPT_PSWD
 readonly WSREP_SST_OPT_AUTH
-
-if [ -n "$WSREP_SST_OPT_REMOTE_AUTH" ]; then
-    # Split auth string at the last ':'
-    readonly WSREP_SST_OPT_REMOTE_USER="${WSREP_SST_OPT_REMOTE_AUTH%%:*}"
-    readonly WSREP_SST_OPT_REMOTE_PSWD="${WSREP_SST_OPT_REMOTE_AUTH#*:}"
-else
-    readonly WSREP_SST_OPT_REMOTE_USER=
-    readonly WSREP_SST_OPT_REMOTE_PSWD=
-fi
-
+readonly WSREP_SST_OPT_REMOTE_USER
+readonly WSREP_SST_OPT_REMOTE_PSWD
 readonly WSREP_SST_OPT_REMOTE_AUTH
 
 if [ -n "$WSREP_SST_OPT_DATA" ]; then
@@ -1015,7 +1144,7 @@ wsrep_check_program()
     local prog="$1"
     local cmd=$(commandex "$prog")
     if [ -z "$cmd" ]; then
-        echo "'$prog' not found in PATH"
+        wsrep_log_error "'$prog' not found in path"
         return 2 # no such file or directory
     fi
 }
@@ -1043,7 +1172,7 @@ get_openssl()
 {
     # If the OPENSSL_BINARY variable is already defined, just return:
     if [ -n "${OPENSSL_BINARY+x}" ]; then
-        return
+        return 0
     fi
     # Let's look for openssl:
     OPENSSL_BINARY=$(commandex 'openssl')
@@ -1068,9 +1197,7 @@ wsrep_gen_secret()
         printf '%04x%04x%04x%04x%04x%04x%04x%04x' \
                $RANDOM $RANDOM $RANDOM $RANDOM \
                $RANDOM $RANDOM $RANDOM $RANDOM
-    elif [ -n "$(commandex cksum)" -a \
-           -n "$(commandex printf)" ]
-    then
+    elif [ $with_printf -ne 0 -a -n "$(commandex cksum)" ]; then
         printf '%08x%08x%08x%08x' \
             $(head -8 /dev/urandom | cksum | cut -d ' ' -f1) \
             $(head -8 /dev/urandom | cksum | cut -d ' ' -f1) \
@@ -1106,9 +1233,9 @@ is_local_ip()
     # the domain name check:
     if [ "${2:-0}" -eq 0 ]; then
        # We consider all the names of a given host to be local addresses:
-       [ "$1" = "$(hostname -s)" -o \
-         "$1" = "$(hostname -f)" -o \
-         "$1" = "$(hostname -d)" ] && return 0
+       [ "$1" = "$(hostname -s 2>/dev/null)" -o \
+         "$1" = "$(hostname -f 2>/dev/null)" -o \
+         "$1" = "$(hostname -d 2>/dev/null)" ] && return 0
     fi
     # If the address contains anything other than digits
     # and separators, it is not a local address:
@@ -1136,22 +1263,56 @@ is_local_ip()
 
 check_sockets_utils()
 {
+    # The presence of any of these utilities is enough for us:
     lsof_available=0
     sockstat_available=0
     ss_available=0
 
-    [ -n "$(commandex lsof)" ] && lsof_available=1
-    [ -n "$(commandex sockstat)" ] && sockstat_available=1
-    [ -n "$(commandex ss)" ] && ss_available=1
-
-    if [ $lsof_available -eq 0 -a \
-         $sockstat_available -eq 0 -a \
-         $ss_available -eq 0 ]
-    then
-        wsrep_log_error "Neither lsof, nor sockstat or ss tool was found in" \
-                        "the PATH. Make sure you have it installed."
-        exit 2 # ENOENT
+    socket_utility="$(commandex ss)"
+    if [ -n "$socket_utility" ]; then
+        socket_uname='ss'
+        ss_available=1
+        ss_opts='-nlp'
+        # Let's check that ss has an option to skip headers:
+        if $socket_utility -h 2>&1 | grep -qw -F -- '-H'; then
+            ss_available=2
+            ss_opts="${ss_opts}H"
+        fi
+    else
+        socket_utility="$(commandex sockstat)"
+        if [ -n "$socket_utility" ]; then
+            socket_uname='sockstat'
+            sockstat_available=1
+            sockstat_opts='-p'
+            if [ "$OS" = 'FreeBSD' ]; then
+                # sockstat in FreeBSD is different from other systems,
+                # let's denote it with a different value:
+                sockstat_available=2
+                sockstat_opts='-46lq -P tcp -p'
+            fi
+        else
+            socket_utility="$(commandex lsof)"
+            if [ -n "$socket_utility" ]; then
+                socket_uname='lsof'
+                lsof_available=1
+                lsof_opts='-Pnl'
+                # Let's check that lsof has an option to bypass blocking:
+                if $socket_utility -h 2>&1 | grep -qw -F -- '-b'; then
+                    lsof_available=2
+                    lsof_opts="$lsof_opts -b -w"
+                else
+                    lsof_opts="$lsof_opts -S 10"
+                fi
+            else
+                wsrep_log_error "Neither lsof, nor sockstat, nor ss tool" \
+                                "were found in the path. Make sure you have" \
+                                "at least one of them installed."
+                exit 2 # ENOENT
+            fi
+        fi
     fi
+    wsrep_log_info "'$socket_uname' is selected as a socket" \
+                   "information utility."
 }
 
 #
@@ -1172,23 +1333,33 @@ check_port()
 
     [ $pid -le 0 ] && pid='[0-9]+'
 
-    local rc=1
+    local rc=2 # ENOENT
 
-    if [ $lsof_available -ne 0 ]; then
-        lsof -Pnl -i ":$port" 2>/dev/null | \
-        grep -q -E "^($utils)[^[:space:]]*[[:space:]]+$pid[[:space:]].*\\(LISTEN\\)" && rc=0
+    if [ $ss_available -ne 0 ]; then
+        $socket_utility $ss_opts -t "( sport = :$port )" 2>/dev/null | \
+            grep -q -E "[[:space:]]users:[[:space:]]?\\(.*\\(\"($utils)[^[:space:]]*\"[^)]*,pid=$pid(,[^)]*)?\\)" && rc=0
     elif [ $sockstat_available -ne 0 ]; then
-        local opts='-p'
-        if [ "$OS" = 'FreeBSD' ]; then
-            # sockstat on FreeBSD requires the "-s" option
-            # to display the connection state:
-            opts='-sp'
+        if [ $sockstat_available -gt 1 ]; then
+            # The sockstat command on FreeBSD does not return
+            # the connection state without special option, but
+            # it supports filtering by connection state:
+            local out
+            out=$($socket_utility $sockstat_opts "$port" 2>/dev/null) || rc=16 # EBUSY
+            # On FreeBSD, the sockstat utility may exit without
+            # any output due to locking issues in certain versions;
+            # let's return a special exit code in such cases:
+            if [ $rc -eq 16 -o -z "$out" ]; then
+                return 16 # EBUSY
+            fi
+            echo "$out" | \
+                grep -q -E "^[^[:space:]]+[[:space:]]+($utils)[^[:space:]]*[[:space:]]+$pid([[:space:]]|\$)" && rc=0
+        else
+            $socket_utility $sockstat_opts "$port" 2>/dev/null | \
+                grep -q -E "^[^[:space:]]+[[:space:]]+($utils)[^[:space:]]*[[:space:]]+$pid([[:space:]].+)?[[:space:]]LISTEN([[:space:]]|\$)" && rc=0
         fi
-        sockstat "$opts" "$port" 2>/dev/null | \
-        grep -q -E "[[:space:]]+($utils)[^[:space:]]*[[:space:]]+$pid[[:space:]].*[[:space:]]LISTEN" && rc=0
-    elif [ $ss_available -ne 0 ]; then
-        ss -nlpH "( sport = :$port )" 2>/dev/null | \
-        grep -q -E "users:\\(.*\\(\"($utils)[^[:space:]]*\"[^)]*,pid=$pid(,[^)]*)?\\)" && rc=0
+    elif [ $lsof_available -ne 0 ]; then
+        $socket_utility $lsof_opts -i ":$port" 2>/dev/null | \
+            grep -q -E "^($utils)[^[:space:]]*[[:space:]]+$pid([[:space:]].+)?[[:space:]]\\(LISTEN\\)([[:space:]]|\$)" && rc=0
     else
         wsrep_log_error "Unknown sockets utility"
         exit 2 # ENOENT
@@ -1271,7 +1442,7 @@ verify_ca_matches_cert()
         wsrep_log_info "run: \"$OPENSSL_BINARY\" verify -verbose${ca:+ -CAfile \"$ca\"}${cap:+ -CApath \"$cap\"} \"$cert\""
         wsrep_log_info "output: $errmsg"
         wsrep_log_error "******** FATAL ERROR ********************************************"
-        wsrep_log_error "* The certifcate and CA (certificate authority) do not match.   *"
+        wsrep_log_error "* The certificate and CA (certificate authority) do not match.  *"
         wsrep_log_error "* It does not appear that the certificate was issued by the CA. *"
         wsrep_log_error "* Please check your certificate and CA files.                   *"
         wsrep_log_error "*****************************************************************"
@@ -1440,7 +1611,7 @@ cleanup_pid()
                     fi
                 done
             elif ps -p $pid >/dev/null 2>&1; then
-                wsrep_log_warning "Unable to kill PID=$pid ($pid_file)"
+                wsrep_log_warning "Unable to kill PID=$pid${pid_file:+ ($pid_file)}"
                 return 1
             fi
         fi
@@ -1459,7 +1630,7 @@ get_proc()
     if [ -z "$nproc" ]; then
         set +e
         if [ "$OS" = 'Linux' ]; then
-            nproc=$(grep -cw -E '^processor' /proc/cpuinfo 2>/dev/null)
+            nproc=$(grep -cw -E '^processor' /proc/cpuinfo 2>/dev/null || :)
         elif [ "$OS" = 'Darwin' -o "$OS" = 'FreeBSD' ]; then
             nproc=$(sysctl -n hw.ncpu)
         fi
@@ -1509,6 +1680,139 @@ check_server_ssl_config()
     fi
 }
 
+# Get Common Name (CN) from the certificate:
+openssl_getCN()
+{
+    get_openssl
+    if [ -z "$OPENSSL_BINARY" ]; then
+        wsrep_log_error \
+            'openssl not found but it is required for authentication'
+        exit 42
+    fi
+
+    local bug=0
+    local CN=$("$OPENSSL_BINARY" x509 -noout -subject -in "$1" 2>&1) || bug=1
+
+    if [ $bug -ne 0 ]; then
+        wsrep_log_info "run: \"$OPENSSL_BINARY\" x509 -noout -subject -in \"$1\""
+        wsrep_log_info "output: $CN"
+        wsrep_log_error "******** FATAL ERROR **********************************************"
+        wsrep_log_error "* Unable to parse the certificate file to obtain the common name. *"
+        wsrep_log_error "*******************************************************************"
+        exit 22
+    fi
+
+    CN=$(trim_string "$CN")
+
+    if [ -n "$CN" ]; then
+        # If the string begins with the "subject" prefix
+        # then we need to remove it:
+        local saved="$CN"
+        local remain="${CN#subject}"
+        if [ "$remain" != "$saved" ]; then
+            remain=$(trim_left "$remain")
+            # Now let's check for the presence of "=" character
+            # after the "subject":
+            saved="$remain"
+            remain="${remain#=}"
+            if [ "$remain" != "$saved" ]; then
+                remain=$(trim_left "$remain")
+            else
+                remain=""
+                bug=1
+            fi
+        fi
+        while [ -n "$remain" ]; do
+            local value=""
+            # Let's extract the option name - all characters
+            # up to the first '=' or ',' character (if present):
+            local option="${remain%%[=,]*}"
+            if [ "$option" != "$remain" ]; then
+                option=$(trim_right "$option")
+                # These variables will be needed to determine
+                # which separator comes first:
+                local x="${remain#*=}"
+                local y="${remain#*,}"
+                local z=${#remain}
+                x=${#x}; [ $x -eq $z ] && x=0
+                y=${#y}; [ $y -eq $z ] && y=0
+                # The remaining string is everything that follows
+                # the separator character:
+                remain=$(trim_left "${remain#*[=,]}")
+                # Let's check what we are dealing with - an equal
+                # sign or a comma?
+                if [ $x -gt $y ]; then
+                    # If the remainder begins with a double quote,
+                    # then there is a string containing commas and
+                    # we need to parse it:
+                    saved="$remain"
+                    remain="${remain#\"}"
+                    if [ "$remain" != "$saved" ]; then
+                        while :; do
+                            # We need to find the closing quote:
+                            local prefix="$remain"
+                            remain="${remain#*\"}"
+                            # Let's check if there is a closing quote?
+                            if [ "$remain" = "$prefix" ]; then
+                                bug=1
+                                break
+                            fi
+                            # Everything up to the closing quote is
+                            # the next part of the value:
+                            value="$value${prefix%%\"*}"
+                            # But if the last character of the value
+                            # is a backslash, then it is a quoted quotation
+                            # mark and we need to add it to the value:
+                            if [ "${value%\\}" != "$value" ]; then
+                                value="$value\""
+                            else
+                                break
+                            fi
+                        done
+                        [ $bug -ne 0 ] && break
+                        # Now we have to remove "," if it is present
+                        # in the string after the value:
+                        saved=$(trim_left "$remain")
+                        remain="${saved#,}"
+                        if [ "$remain" != "$saved" ]; then
+                            remain=$(trim_left "$remain")
+                        elif [ -n "$remain" ]; then
+                            bug=1
+                            break
+                        fi
+                    else
+                        # We are dealing with a simple unquoted string value,
+                        # therefore we need to take everything up to the end
+                        # of the string, or up to the next comma character:
+                        value="${remain%%,*}"
+                        if [ "$value" != "$remain" ]; then
+                            remain=$(trim_left "${remain#*,}")
+                        else
+                            remain=""
+                        fi
+                        value=$(trim_right "$value")
+                    fi
+                    if [ "$option" = 'CN' -a -n "$value" ]; then
+                        echo "$value"
+                        return
+                    fi
+                fi
+            else
+                remain=""
+            fi
+        done
+    fi
+
+    if [ $bug -ne 0 ]; then
+        wsrep_log_error "******** FATAL ERROR **********************************************"
+        wsrep_log_error "* Unable to parse the certificate options: '$CN'"
+        wsrep_log_error "*******************************************************************"
+        exit 22
+    fi
+
+    echo ''
+}
+
 simple_cleanup()
 {
     # Since this is invoked just after exit NNN
@@ -1516,11 +1820,144 @@ simple_cleanup()
     if [ $estatus -ne 0 ]; then
         wsrep_log_error "Cleanup after exit with status: $estatus"
     fi
-    if [ -n "${SST_PID:-}" ]; then
+    if [ -n "$SST_PID" ]; then
         [ "$(pwd)" != "$OLD_PWD" ] && cd "$OLD_PWD"
         [ -f "$SST_PID" ] && rm -f "$SST_PID" || :
     fi
     exit $estatus
 }
 
+create_data()
+{
+    OLD_PWD="$(pwd)"
+    DATA_DIR="$OLD_PWD"
+
+    if [ -n "$DATA" -a "$DATA" != '.' ]; then
+        [ ! -d "$DATA" ] && mkdir -p "$DATA"
+        cd "$DATA"
+        DATA_DIR="$(pwd)"
+        cd "$OLD_PWD"
+    fi
+}
+
+create_dirs()
+{
+    local simplify=${1:-0}
+
+    # if no command line argument and INNODB_DATA_HOME_DIR environment
+    # variable is not set, try to get it from the my.cnf:
+    if [ -z "$INNODB_DATA_HOME_DIR" ]; then
+        INNODB_DATA_HOME_DIR=$(parse_cnf '--mysqld' 'innodb-data-home-dir')
+        INNODB_DATA_HOME_DIR=$(trim_dir "$INNODB_DATA_HOME_DIR")
+    fi
+
+    if [ -n "$INNODB_DATA_HOME_DIR" -a "$INNODB_DATA_HOME_DIR" != '.' -a \
+         "$INNODB_DATA_HOME_DIR" != "$DATA_DIR" ]
+    then
+        # handle both relative and absolute paths:
+        cd "$DATA"
+        [ ! -d "$INNODB_DATA_HOME_DIR" ] && mkdir -p "$INNODB_DATA_HOME_DIR"
+        cd "$INNODB_DATA_HOME_DIR"
+        ib_home_dir="$(pwd)"
+        cd "$OLD_PWD"
+        [ $simplify -ne 0 -a "$ib_home_dir" = "$DATA_DIR" ] && ib_home_dir=""
+    fi
+
+    # if no command line argument and INNODB_LOG_GROUP_HOME is not set,
+    # then try to get it from the my.cnf:
+    if [ -z "$INNODB_LOG_GROUP_HOME" ]; then
+        INNODB_LOG_GROUP_HOME=$(parse_cnf '--mysqld' 'innodb-log-group-home-dir')
+        INNODB_LOG_GROUP_HOME=$(trim_dir "$INNODB_LOG_GROUP_HOME")
+    fi
+
+    if [ -n "$INNODB_LOG_GROUP_HOME" -a "$INNODB_LOG_GROUP_HOME" != '.' -a \
+         "$INNODB_LOG_GROUP_HOME" != "$DATA_DIR" ]
+    then
+        # handle both relative and absolute paths:
+        cd "$DATA"
+        [ ! -d "$INNODB_LOG_GROUP_HOME" ] && mkdir -p "$INNODB_LOG_GROUP_HOME"
+        cd "$INNODB_LOG_GROUP_HOME"
+        ib_log_dir="$(pwd)"
+        cd "$OLD_PWD"
+        [ $simplify -ne 0 -a "$ib_log_dir" = "$DATA_DIR" ] && ib_log_dir=""
+    fi
+
+    # if no command line argument and INNODB_UNDO_DIR is not set,
+    # then try to get it from the my.cnf:
+    if [ -z "$INNODB_UNDO_DIR" ]; then
+        INNODB_UNDO_DIR=$(parse_cnf '--mysqld' 'innodb-undo-directory')
+        INNODB_UNDO_DIR=$(trim_dir "$INNODB_UNDO_DIR")
+    fi
+
+    if [ -n "$INNODB_UNDO_DIR" -a "$INNODB_UNDO_DIR" != '.' -a \
+         "$INNODB_UNDO_DIR" != "$DATA_DIR" ]
+    then
+        # handle both relative and absolute paths:
+        cd "$DATA"
+        [ ! -d "$INNODB_UNDO_DIR" ] && mkdir -p "$INNODB_UNDO_DIR"
+        cd "$INNODB_UNDO_DIR"
+        ib_undo_dir="$(pwd)"
+        cd "$OLD_PWD"
+        [ $simplify -ne 0 -a "$ib_undo_dir" = "$DATA_DIR" ] && ib_undo_dir=""
+    fi
+
+    # if no command line argument then try to get it from the my.cnf:
+    if [ -z "$ARIA_LOG_DIR" ]; then
+        ARIA_LOG_DIR=$(parse_cnf '--mysqld' 'aria-log-dir-path')
+        ARIA_LOG_DIR=$(trim_dir "$ARIA_LOG_DIR")
+    fi
+
+    if [ -n "$ARIA_LOG_DIR" -a "$ARIA_LOG_DIR" != '.' -a \
+         "$ARIA_LOG_DIR" != "$DATA_DIR" ]
+    then
+        # handle both relative and absolute paths:
+        cd "$DATA"
+        [ ! -d "$ARIA_LOG_DIR" ] && mkdir -p "$ARIA_LOG_DIR"
+        cd "$ARIA_LOG_DIR"
+        ar_log_dir="$(pwd)"
+        cd "$OLD_PWD"
+        [ $simplify -ne 0 -a "$ar_log_dir" = "$DATA_DIR" ] && ar_log_dir=""
+    fi
+
+    return 0
+}
+
+wait_previous_sst()
+{
+    # give some time for previous SST to complete:
+    check_round=0
+    while check_pid "$SST_PID" 1; do
+        wsrep_log_info "Previous SST is not completed, waiting for it to exit"
+        check_round=$(( check_round+1 ))
+        if [ $check_round -eq 30 ]; then
+            wsrep_log_error "previous SST script still running..."
+            exit 114 # EALREADY
+        fi
+        sleep 1
+    done
+
+    trap simple_cleanup EXIT
+    echo $$ > "$SST_PID"
+}
+
+DATA="$WSREP_SST_OPT_DATA"
+
+wsrep_check_datadir
+create_data
+
+SST_PID="$DATA/wsrep_sst.pid"
+
+if [ -n "${MTR_SST_JOINER_DELAY:-}" ]; then
+    MTR_SST_JOINER_DELAY=$(trim_string "$MTR_SST_JOINER_DELAY")
+fi
+
+simulate_long_sst()
+{
+    # Delay for MTR tests if needed to simulate long SST/IST:
+    if [ ${MTR_SST_JOINER_DELAY:-0} -gt 0 ]; then
+        wsrep_log_info "Sleeping $MTR_SST_JOINER_DELAY seconds for MTR test"
+        sleep $MTR_SST_JOINER_DELAY
+    fi
+}
+
 wsrep_log_info "$WSREP_METHOD $WSREP_TRANSFER_TYPE started on $WSREP_SST_OPT_ROLE"
